<?php

namespace Drupal\h5p_challenge\Controller;

use Drupal\Core\Controller\ControllerBase;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Database\Driver\mysql\Connection;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\JsonResponse;
use Drupal\Component\Uuid\Php as Uuid;
use Drupal\h5p_challenge\H5PChallengeService;
use Symfony\Component\HttpFoundation\Request;
use Drupal\Core\Datetime\DateFormatter;
use Drupal\Core\Mail\MailManager;
use Drupal\Core\Url;
use Drupal\h5p_challenge\FetchClass\H5PChallenge;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Drupal\Core\Link;
use Drupal\language\ConfigurableLanguageManager;

/**
 * Class H5PChallengeController.
 */
class H5PChallengeController extends ControllerBase {

  /**
   * Drupal\Core\Database\Driver\mysql\Connection definition.
   *
   * @var \Drupal\Core\Database\Driver\mysql\Connection
   */
  protected $database;

  /**
   * Uuid service
   *
   * @var \Drupal\Component\Uuid\Php
   */
  protected $uuid;

  /**
   * Date formatter
   *
   * @var \Drupal\Core\Datetime\DateFormatter
   */
  protected $dateFormatter;

  /**
   * Mail manager
   *
   * @var \Drupal\Core\Mail\MailManager
   */
  protected $mailManager;

  /**
   * Challenge service
   *
   * @var \Drupal\h5p_challenge\H5PChallengeService
   */
  protected $challengeService;

  /**
   * Language manager
   *
   * @var \Drupal\language\ConfigurableLanguageManager
   */
  protected $languageManager;

  /**
   * Constructs a new H5PChallengeController object.
   */
  public function __construct(Connection $database, Uuid $uuid, H5PChallengeService $challengeService, DateFormatter $dateFormatter, MailManager $mailManager, ConfigurableLanguageManager $languageManager) {
    $this->database = $database;
    $this->uuid = $uuid;
    $this->challengeService = $challengeService;
    $this->dateFormatter = $dateFormatter;
    $this->mailManager = $mailManager;
    $this->languageManager = $languageManager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('database'),
      $container->get('uuid'),
      $container->get('h5p.challenge.default'),
      $container->get('date.formatter'),
      $container->get('plugin.manager.mail'),
      $container->get('language_manager')
    );
  }

  /**
   * Start new challenge
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   JSON response with possible additional data or message
   */
  public function createNew(Request $request) {
    $content_id = filter_input(INPUT_POST, 'contentId', FILTER_VALIDATE_INT);
    $title = filter_input(INPUT_POST, 'title');
    $email = filter_input(INPUT_POST, 'email', FILTER_VALIDATE_EMAIL);
    $duration = filter_input(INPUT_POST, 'duration', FILTER_VALIDATE_INT);
    $captcha = filter_input(INPUT_POST, 'g-recaptcha-response');
    $token = filter_input(INPUT_GET, 'token');

    if (! ($content_id && $title && $email !== NULL && $email !== FALSE && $duration !== NULL && $duration !== FALSE && $captcha && $token)) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Wrong usage'),
      ]);
    }

    /*if (! \H5PCore::validToken('result', $token)) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Invalid security token'),
      ]);
    }*/

    if (! \Drupal::service('email.validator')->isValid($email)) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Provided email address is invalid'),
      ]);
    }

    if ($duration < 1 && $duration > 168) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Provided duration is invalid'),
      ]);
    }

    if (!$this->challengeService->checkIfContentExists($content_id)) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Content does not exist'),
      ]);
    }

    try {
      $captchaResponse = \Drupal::httpClient()->post(H5P_CHALLENGE_RECAPTCHA_VERIFY_ENDPOINT, [
        'form_params' => [
          'secret' => $this->challengeService->getRecaptchaSecretKey(),
          'response' => $captcha,
          'remoteip' => $request->getClientIp(),
        ],
      ]);
    } catch (\Exception $e) {
      \Drupal::logger('h5p_challenge')->error($e->getMessage());

      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Captcha service responded with an error'),
      ]);
    }

    if ($captchaResponse->getStatusCode() !== 200 || $captchaResponse->getReasonPhrase() !== 'OK') {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Captcha service responded with an error'),
      ]);
    }

    $parsedResponse = json_decode($captchaResponse->getBody(), TRUE);
    if ($parsedResponse['success'] !== TRUE) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Wrong or unsuitable captcha'),
      ]);
    }

    $started = time();
    $finished = $started + ($duration * 60 * 60);
    $code = $this->challengeService->generateUniqueChallengeCode($content_id);
    $langcode = $this->languageManager->getCurrentLanguage()->getId();
    $uuid = $this->uuid->generate();
    $url = Url::fromRoute('h5p_challenge.h5p_challenge_controller_results', [
      'uuid' => $uuid,
    ], [
      'absolute' => TRUE,
    ])->toString();

    // TODO Handle failures?
    $this->database->insert('h5p_challenge')->fields([
      'uuid' => $uuid,
      'content_id' => $content_id,
      'title' => $title,
      'email' => $email,
      'started' => $started,
      'finished' => $finished,
      'code' => $code,
      'langcode' => $langcode,
    ])->execute();

    $message = t("You have created a new challenge: @title.\nThe title of a resource for this challenge is: @ctitle.\n\nThe challenge will be running from @started to @finished (the duration would be @hours hours).\n\nChallenge code to participate is: @code\n\nYou can track results at: @results\nAnother notification containing results would be sent to you soon after the challenge is over.\nThe results link would still be active for at least 36 hours after the challenge ends.\n\n\nBest regards,\nChallenge team.", [
      '@title' => $title,
      '@ctitle' => H5PChallenge::getContentTitleById($content_id),
      '@started' => $this->dateFormatter->format($started, 'long'),
      '@finished' => $this->dateFormatter->format($finished, 'long'),
      '@hours' => $duration,
      '@code' => $code,
      '@results' => $url,
    ]);
    $params = [
      'body' => (string) $message,
    ];
    $this->mailManager->mail('h5p_challenge', 'new_challenge_created', $email, $langcode, $params, NULL, TRUE);

    return new JsonResponse([
      'success' => TRUE,
      'code' => $code,
      'url' => $url,
    ]);
  }

  /**
   * Join existing challenge
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   JSON response with possible additional data or message
   */
  public function startPlaying() {
    $content_id = filter_input(INPUT_POST, 'contentId', FILTER_VALIDATE_INT);
    $name = filter_input(INPUT_POST, 'name');
    $code = filter_input(INPUT_POST, 'code');
    $token = filter_input(INPUT_GET, 'token');

    if (! ($content_id && $name && $code && $token)) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Wrong usage'),
      ]);
    }

    /*if (! \H5PCore::validToken('result', $token)) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Invalid security token'),
      ]);
    }*/

    $challenge = $this->challengeService->getChallengeByContentIdAndCode($content_id, $code);

    if ($challenge && $challenge->isActive()) {
      $uuid = $this->uuid->generate();

      $this->database->insert('h5p_challenge_points')->fields([
        'challenge_uuid' => $challenge->uuid,
        'player_uuid' => $uuid,
        'name' => $name,
        'started' => time(),
      ])->execute();

      $data = [
        'uuid' => $uuid,
        'challenge' => [
          'title' => $challenge->title,
          'started' => $challenge->started,
          'finished' => $challenge->finished,
        ],
      ];

      user_cookie_save(['h5p-challenge-for-' . $content_id => json_encode($data),]);

      return new JsonResponse(array_merge($data, [
        'success' => TRUE,
      ]));
    } else if ($challenge) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Challenge is no longer active'),
      ]);
    } else {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('No challenge could be found for provided code'),
      ]);
    }

    return new JsonResponse([
      'success' => FALSE,
    ]);
  }

  /**
   * Store results for challenge participant
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   JSON response with possible additional data or message
   */
  public function setFinished() {
    $content_id = filter_input(INPUT_POST, 'contentId', FILTER_VALIDATE_INT);
    $uuid = filter_input(INPUT_POST, 'uuid');
    $score = filter_input(INPUT_POST, 'score', FILTER_VALIDATE_INT);
    $max_score = filter_input(INPUT_POST, 'maxScore', FILTER_VALIDATE_INT);
    $token = filter_input(INPUT_GET, 'token');

    if (! ($content_id && $uuid && $score !== NULL && $score !== FALSE && $max_score !== NULL && $max_score !== FALSE && $token)) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Wrong usage'),
      ]);
    }

    /*if (! \H5PCore::validToken('result', $token)) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Invalid security token'),
      ]);
    }*/

    $points = $this->challengeService->getChallengePointsByUuid($uuid);

    if ($points) {
      $challenge = $this->challengeService->getChallengeByUuid($points->challenge_uuid);

      if ($challenge && $challenge->isActive()) {
        $finished_time = time();

        $this->database->update('h5p_challenge_points')
        ->fields([
          'finished' => $finished_time,
          'points' => $score,
          'max_points' => $max_score,
        ])
        ->expression('tries_count', 'tries_count + :amount', [':amount' => 1,])
        ->condition('player_uuid', $uuid)
        ->execute();

        // Update cookie contents to indicate that score has been stored
        $cname = 'Drupal_visitor_h5p-challenge-for-' . $content_id;
        if (isset($_COOKIE[$cname])) {
          $data = json_decode($_COOKIE[$cname], TRUE);
          $data['finished'] = [
            'finished' => $finished_time,
            'score' => $score,
            'maxScore' => $max_score,
          ];
          user_cookie_save(['h5p-challenge-for-' . $content_id => json_encode($data),]);
        }

        return new JsonResponse([
          'success' => TRUE,
        ]);
      } else if ($challenge && !$challenge->isActive()) {
        return new JsonResponse([
          'success' => FALSE,
          'message' => $this->t('Challenge is no longer active'),
        ]);
      }
    }

    return new JsonResponse([
      'success' => FALSE,
    ]);
  }

  /**
   * Show challenge results page
   *
   * @return array
   *   Page structure
   */
  public function results(Request $request) {
    $challenge = $this->challengeService->getChallengeByUuid($request->attributes->get('uuid'));

    if ($challenge) {
      $response = [];
      $response['#title'] = $this->t('Challenge results: @title', [
        '@title' => $challenge->title,
      ]);

      $header = [$this->t('Name'), $this->t('Started'), $this->t('Finished'), $this->t('Points'), $this->t('Maximum Points'), $this->t('Tries'),];
      $rows = [];

      $result = $this->database->select('h5p_challenge_points', 'cp')
      ->fields('cp')
      ->condition('cp.challenge_uuid', $challenge->uuid)
      ->execute();

      if ($result) {
        while($single = $result->fetchObject()) {
          $rows[] = [
            'data' => [
              'name' => $single->name,
              'started' => $this->dateFormatter->format($single->started, 'long'),
              'finished' => $single->finished ? $this->dateFormatter->format($single->finished, 'long') : '',
              'points' => isset($single->points) ? $single->points : '',
              'max_points' => isset($single->max_points) ? $single->max_points : '',
              'tries_count' => $single->tries_count,
            ],
          ];
        }
      }

      $response['results-csv'] = [
        '#type' => 'link',
        '#title' => $this->t('Download results as CSV'),
        '#url' => Url::fromRoute('h5p_challenge.h5p_challenge_controller_results_csv', [
          'uuid' => $challenge->uuid,
        ], [
          'absolute' => TRUE,
        ]),
        '#attributes' => [
          'class' => ['btn', 'btn-primary', 'btn-xs',],
        ],
      ];

      $response['results-table'] = [
        '#theme' => 'table',
        '#rows' => $rows,
        '#header' => $header,
      ];

      return $response;
    }

    throw new NotFoundHttpException();
  }

  /**
   * Serve challenge results to be downloaded as CSV
   *
   * @return \Symfony\Component\HttpFoundation\Result
   *   Challenge results file
   */
  public function resultsCsv(Request $request) {
    $challenge = $this->challengeService->getChallengeByUuid($request->attributes->get('uuid'));

    if ($challenge) {
      $response = new Response($this->challengeService->getChallengeReultsCsv($challenge));

      $response->headers->set('Content-Type', 'text/csv');
      $response->headers->set('Content-Disposition','attachment; filename="results.csv"');

      return $response;
    }

    throw new NotFoundHttpException();
  }

  /**
   * List of currrently present challenges
   *
   * @return array
   *   Page structure
   */
  public function list() {
    $query = $this->database->select('h5p_challenge', 'c', [
      'fetch' => H5PChallenge::class,
    ]);
    $query->fields('c', ['uuid', 'content_id', 'title', 'email', 'started', 'finished', 'code', 'results_sent', 'langcode',]);
    $pager = $query->extend('Drupal\Core\Database\Query\PagerSelectExtender')->limit(25);
    $result = $pager->execute();

    $header = [
      $this->t('Title'),
      $this->t('Started'),
      $this->t('Finished'),
      $this->t('Code'),
      $this->t('Results sent'),
      $this->t('Language'),
    ];
    $rows = [];

    foreach ($result->fetchAll() as $challenge) {
      $rows[] = [
        'data' => [
          'title' => Link::createFromRoute(
            $challenge->title,
            'h5p_challenge.h5p_challenge_controller_results',
            ['uuid' => $challenge->uuid],
            [
              'attributes' => [
                'target' => '_blank',
              ]
            ]
          ),
          'started' => $this->dateFormatter->format($challenge->started, 'long'),
          'finished' => $this->dateFormatter->format($challenge->finished, 'long'),
          'code' => $challenge->code,
          'results_sent' => ($challenge->results_sent) ? $this->t('Yes') : $this->t('No'),
          'language' => $this->languageManager->getLanguageName($challenge->langcode),
        ],
      ];
    }

    $response['challenges'] = [
      '#type' => 'table',
      '#caption' => $this->t('Challenges'),
      '#header' => $header,
      '#rows' => $rows,
      '#empty' => $this->t('No challenges found'),
    ];

    $response['pager'] = [
      '#type' => 'pager',
    ];

    return $response;
  }

}
