<?php

namespace Drupal\h5p_challenge;
use Drupal\Core\Database\Driver\mysql\Connection;
use Drupal\Core\Config\ConfigFactory;
use Drupal\h5p_challenge\FetchClass\H5PChallenge;
use Drupal\Core\Url;
use Drupal\Core\Datetime\DateFormatter;
use Drupal\Core\Mail\MailManager;

/**
 * Class H5PChallengeService.
 */
class H5PChallengeService {

  /**
   * Drupal\Core\Database\Driver\mysql\Connection definition.
   *
   * @var \Drupal\Core\Database\Driver\mysql\Connection
   */
  protected $database;

  /**
   * Drupal\Core\Config\ConfigFactory definition.
   *
   * @var \Drupal\Core\Config\ConfigFactory
   */
  protected $configFactory;

  /**
   * Date formatter
   *
   * @var \Drupal\Core\Datetime\DateFormatter
   */
  protected $dateFormatter;

  /**
   * Mail manager
   *
   * @var \Drupal\Core\Mail\MailManager
   */
  protected $mailManager;

  /**
   * Constructs a new H5PChallengeService object.
   */
  public function __construct(Connection $database, ConfigFactory $configFactory, DateFormatter $dateFormatter, MailManager $mailManager) {
    $this->database = $database;
    $this->configFactory = $configFactory;
    $this->dateFormatter = $dateFormatter;
    $this->mailManager = $mailManager;
  }

  /**
   * Returns reCaptcha site key setting value
   * @return string
   *   Site key setting
   */
  public function getRecaptchaSiteKey() {
    $config = $this->configFactory->get('h5p_challenge.config');

    return $config->get('recaptcha_site_key');
  }

  /**
   * Returns reCaptcha secret key setting value
   * @return string
   *   Secret key setting
   */
  public function getRecaptchaSecretKey() {
    $config = $this->configFactory->get('h5p_challenge.config');

    return $config->get('recaptcha_secret_key');
  }

  /**
   * Returns email address to be used when sending notifications
   * @return string
   *   Email address setting
   */
  public function getEmailFromAddress() {
    $config = $this->configFactory->get('h5p_challenge.config');

    return $config->get('email_from_address');
  }

  /**
   * Generates a string of random numbers from 0 to 9 for a given length
   * Source: https://stackoverflow.com/a/13169091/2704169
   * @param  int    $length
   *   Length of the result string
   * @return string
   *   String of random numbers with given length
   */
  public function generateRandomNumber(int $length) {
    $result = '';

    for($i = 0; $i < $length; $i++) {
      $result .= (function_exists('random_int')) ? random_int(0, 9) : mt_rand(0, 9);
    }

    return $result;
  }

  /**
   * Makes sure that code is unique for certain content
   * @param  string  $code
   *   Code to check
   * @param  int     $contentId
   *   Content identifier
   * @return boolean
   *   Code is not yet being used
   */
  public function isUniqueChallengeCode(string $code, int $contentId) {
    $query = $this->database->select('h5p_challenge', 'c');
    $query->condition('c.content_id', $contentId);
    $query->condition('c.code', $code);
    $query->addExpression('COUNT(*)');
    $query->countQuery();
    $count = $query->execute()->fetchField();

    return (int)$count === 0;
  }

  /**
   * Generates unique challenge code for certain content
   * @param  int    $contentId
   *   Content identifier
   * @return string
   *   Unique numeric code
   */
  public function generateUniqueChallengeCode(int $contentId) {
    $iteration = 1;
    $length = 6;

    $code = $this->generateRandomNumber($length);

    while(!$this->isUniqueChallengeCode($code, $contentId)) {
      $code = $this->generateRandomNumber($length);
      $iteration++;
      if ($iteration > 3) {
        $length++;
        $iteration = 0;
      }
    }

    return $code;
  }

  /**
   * Determines if content with provided identifier does exist
   * @param  int    $contentId
   *   Content identifier
   * @return boolean
   *   Content exists or not
   */
  public function checkIfContentExists(int $contentId) {
    $content = $this->database->select('h5p_content', 'c')
    ->fields('c', ['id',])
    ->condition('c.id', $contentId)
    ->execute()
    ->fetchObject();

    return !!$content;
  }

  /**
   * Loads challenge by content id and code
   * @param  int    $contentId
   *   Content identifier the challenge is created for
   * @param  string $code
   *   Challenge code that is unique in context of certain content
   * @return \Drupal\h5p_challenge\FetchClass\H5PChallenge
   *   Challenge instance or empty result
   */
  public function getChallengeByContentIdAndCode(int $contentId, string $code) {
    $result = $this->database->select('h5p_challenge', 'c', [
      'fetch' => H5PChallenge::class,
    ])
    ->fields('c')
    ->condition('c.content_id', $contentId)
    ->condition('c.code', $code)
    ->execute()
    ->fetch();

    return $result;
  }

  /**
   * Loads challenge points by player unique identifier
   * @param  string $uuid
   *   Player unique identifier
   * @return object
   *   Object or empty result
   */
  public function getChallengePointsByUuid(string $uuid) {
    $result = $this->database->select('h5p_challenge_points', 'cp')
    ->fields('cp')
    ->condition('cp.player_uuid', $uuid)
    ->execute()
    ->fetchObject();

    return $result;
  }

  /**
   * Loads challenge by uique identifier
   * @param  string $uuid
   *   Challenge unique identifier
   * @return \Drupal\h5p_challenge\FetchClass\H5PChallenge
   *   Challenge instance or empty result
   */
  public function getChallengeByUuid(string $uuid) {
    $result = $this->database->select('h5p_challenge', 'c',[
      'fetch' => H5PChallenge::class,
    ])
    ->fields('c')
    ->condition('c.uuid', $uuid)
    ->execute()
    ->fetch();

    return $result;
  }

  /**
   * Returns CSV results for a challenge
   * @param  \Drupal\h5p_challenge\FetchClass\H5PChallenge $challenge
   *   Challenge instance
   * @return string
   *   Results CSV
   */
  public function getChallengeReultsCsv(H5PChallenge $challenge) {
    $fp = fopen('php://temp', 'w');
    fputcsv($fp, [t('Name'), t('Started'), t('Finished'), t('Points'), t('Maximum Points'), t('Tries'),]);
    $result = $this->database->select('h5p_challenge_points', 'cp')
    ->fields('cp')
    ->condition('cp.challenge_uuid', $challenge->uuid)
    ->execute();

    if ($result) {
      while($single = $result->fetchObject()) {
        fputcsv($fp, [
          'name' => $single->name,
          'started' => $this->dateFormatter->format($single->started, 'long'),
          'finished' => $single->finished ? $this->dateFormatter->format($single->finished, 'long') : '',
          'points' => isset($single->points) ? $single->points : '',
          'max_points' => isset($single->max_points) ? $single->max_points : '',
          'tries_count' => $single->tries_count,
        ]);
      }
    }

    rewind($fp);
    $filecontent = stream_get_contents($fp);
    fclose($fp);

    return $filecontent;
  }

  /**
   * Retruns count of players that took part in a challenge
   * @return int
   *   Count of players
   */
  private function getChallengePlayerCount(H5PChallenge $challenge) {
    $query = $this->database->select('h5p_challenge_points', 'cp');
    $query->condition('cp.challenge_uuid', $challenge->uuid);
    $query->addExpression('COUNT(*)');
    $query->countQuery();
    return (int)$query->execute()->fetchField();
  }

  /**
   * Determines challenges with unsent results, sends notifications and marks
   * challenge with a specific flag to indicate it.
   */
  public function notificationsCron() {
    $result = $this->database->select('h5p_challenge', 'c', [
      'fetch' => H5PChallenge::class,
    ])
    ->fields('c')
    ->condition('c.results_sent', 0)
    ->condition('c.finished', time(), '<')
    ->execute();

    if ($result) {
      while($challenge = $result->fetch()) {
        $url = Url::fromRoute('h5p_challenge.h5p_challenge_controller_results', [
          'uuid' => $challenge->uuid,
        ], [
          'absolute' => TRUE,
        ])->toString();
        $csv = Url::fromRoute('h5p_challenge.h5p_challenge_controller_results_csv', [
          'uuid' => $challenge->uuid,
        ], [
          'absolute' => TRUE,
        ])->toString();
        $message = t("One of your challenges has ended: @title.\nThe title of a resource for this challenge is: @ctitle.\n\nThe challege was active from @started to @finished.\n\nYou can access results at: @results\nOr download the CSV file at: @csv\n\nThe results link would still be active for at least 36 hours after the challenge ends.\n\n\nBest regards,\nChallenge team.\n\n\n", [
          '@title' => $challenge->title,
          '@ctitle' => $challenge->getContentTitle(),
          '@started' => $this->dateFormatter->format($challenge->started, 'long'),
          '@finished' => $this->dateFormatter->format($challenge->finished, 'long'),
          '@results' => $url,
          '@csv' => $csv,
        ]);

        $params = [
          'body' => (string) $message,
          'attachment' => [
            'filecontent' => $this->getChallengeReultsCsv($challenge),
            'filename' => 'results.csv',
            'filemime' => 'text/csv',
          ],
        ];
        $this->mailManager->mail('h5p_challenge', 'challenge_ended', $challenge->email, $langcode, $params, NULL, TRUE);
        $this->database->update('h5p_challenge')
        ->fields([
          'results_sent' => 1,
        ])
        ->condition('uuid', $challenge->uuid)
        ->execute();

        $this->database->insert('h5p_challenge_statistics')
        ->fields([
          'content_id' => $challenge->content_id,
          'started' => $challenge->started,
          'finished' => $challenge->finished,
          'player_count' => $this->getChallengePlayerCount($challenge),
        ])
        ->execute();
      }
    }
  }

  /**
   * Removes challenges and their results in case more than 36 hours has passed
   * since challenge end time.
   */
  public function cleanupCron() {
    $ids = $this->database->select('h5p_challenge', 'c')
    ->fields('c', ['uuid',])
    ->condition('c.finished', time() - (36 * 60 * 60), '<')
    ->execute()
    ->fetchCol('uuid');

    if ($ids && is_array($ids) && count($ids) > 0) {
      db_delete('h5p_challenge_points')
      ->condition('challenge_uuid', $ids, 'IN')
      ->execute();
      db_delete('h5p_challenge')
      ->condition('uuid', $ids, 'IN')
      ->execute();
    }
  }

  /**
   * Challenge localization
   * @return array
   *   Challenge specific localizations
   */
  public function getLocalization() {
    return [
      'challenge' => t('Challenge'),
      'challengeDescription' => t('Create new or join an existing one'),
      'headerChallenge' => t('Challenge'),
      'buttonJoin' => t('Join (student)'),
      'buttonCreate' => t('Create (teacher)'),
      'buttonGenerateCode' => t('Generate code'),
      'buttonStart' => t('Start'),
      'buttonFinish' => t('Finish'),
      'labelTitle' => t('Title'),
      'labelTeacherEmail' => t('Teacher email'),
      'labelChallengeDuration' => t('Duration'),
      'labelName' => t('Name'),
      'labelCodeInput' => t('Enter code'),
      'placeholderName' => t('Name'),
      'placeholderTitle' => t('Title'),
      'placeholderEmail' => t('some.one@some.provider'),
      'optionOneHour' => t('One hour'),
      'optionTwoHours' => t('Two hours'),
      'optionOneDay' => t('One day'),
      'optionOneWeek' => t('One week'),
      'textChallengeExplanation' => t('NB! Challenge results will be automatically stored after you answer all the questions.'),
      'textChallengeTitle' => t('Challenge: @title'),
      'textChallengeStarted' => t('Challenge started: @date'),
      'textChallengeEnds' => t('Until challenge ends: @timer'),
      'confirmEndChallenge' => t('Clicking "Finish" button will end the Challenge. Score will be stored only if all the questions are answered.'),
      'confirmEndChallengeCanGetScore' => t('Clicking "Finish" button will end the Challenge. We will try to store the current score, provided that one is available.'),
      'errorCouldNotJoinTheChallenge' => t('Something went wrong! Challenge could not be joined.'),
      'errorCouldNotStartNewChallenge' => t('Something went wrong! Challenge could not be created.'),
      'errorCouldNotEndChallenge' => t('Something went wrong! Could not finish the Challenge.'),
      'errorUnknown' => t('Service error! Please try again or contact the administrator.'),
      'successScoreSubmitted' => t('You have successfully completed the Challenge. You could try again or click the "Finish" button.'),
      'challengeUrl' => t('Results URL: @url'),
      'textCheckHint' => t('Please make sure that current content is able to store results for Challenge participants. Try it out for yourself to be sure.'),
    ];
  }

}
